package com.blazinggames.urcl.ui.sprite;

// Java imports
import java.awt.*;

/**
 * Ultimate Retro Common Library
 * Copyright (C) 2001, 2003-2004 Blazing Games Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ------------------------------------------------
 * 
 * The generic sprite class. 
 * 
 * @author Billy D. Spelchan
 * @version OSR1.00
 */
public class Sprite
{
// *********************
// ***** CONSTANTS *****
// *********************

    public static final Point NO_OFFSET = new Point(0,0);
    

// *********************
// ***** VARIABLES *****
// *********************

    protected int _id;
    protected int _layer;

    protected Rectangle _bounds;

    protected Rectangle _dirty;

    protected boolean _isDirty;
    protected boolean _isVisible;
    protected boolean _isSolid;
    protected boolean _isBounded;
    
    protected Color _background;
    protected Color _bounding;


// ***********************
// ***** CONSTRUCTOR *****
// ***********************

    /**
     * Creates an empty sprite
    */
    public Sprite()
    {
        _id = 0;
    	_layer = 0;

        _bounds = new Rectangle(0,0,0,0);
        _dirty = new Rectangle(0,0,0,0);

    	_isDirty = false;
    	_isVisible = false;
    	_isSolid = true;
    	_isBounded = false;
    
    	_background = Color.black;
    	_bounding = Color.red;
    }
    

// **************************************************
// ***** METHODS - Information and modification *****
// **************************************************

    /**
     * sets the sprites ID. for sprite manager use!!!
     *
     * @param n new ID for sprite
    */
    public void setID(int n)
    {
        _id = n;
    }
    
    /**
     * @return ID sprite is currently set for.
    */
    public int getID()
    {
        return _id;
    }
    
    /**
     * sets the layer the sprite is on.
     * THIS SHOULD ONLY BE CALLED BY THE SPRITE MANAGER!!!
     *
     * @param n layer sprite is on
    */
    public void setLayer(int n)
    {
        _layer = n;
    }
    
    /**
     * @return the layer the sprite is on
    */
    public int getLayer()
    {
        return _layer;
    }
    
    /**
     * Changes the location of the sprite.
     *
     * @param x new x location
     * @param y new y location
    */
    public void move(int x, int y)
    {
        if (_isVisible)
            addDirty(_bounds);
        _bounds.setLocation(x, y);
        if (_isVisible)
            addDirty(_bounds);
    }
    
    /**
     * Scales the sprite. Note that not all sprites will support scaling.
     *
     * @param w width to scale to (should be positive)
     * @param h height to scale to (should be positive)
    */
    public void scale(int w, int h)
    {
        if (_isVisible)
            addDirty(_bounds);
        _bounds.setSize(w, h);
        if (_isVisible)
            addDirty(_bounds);
    }
    
    /**
     * Sets the size and placement of the sprite. 
     * Not all sprites will support changing size.
     *
     * @param x New x location of the sprite
     * @param y New y location of the sprite
     * @param w new width of the sprite (some sprites govern their width)
     * @param h new height of the sprite ( some sprites govern their height)
    */
    public void setBounds(int x, int y, int w, int h)
    {
        if (_isVisible)
            addDirty(_bounds);
        _bounds.setBounds(x, y, w, h);
        if (_isVisible)
            addDirty(_bounds);
    }
    
    /**
     * Sets the size and placement of the sprite. 
     * Not all sprites will support changing size.
     *
     * @param r New bounds for the sprite
    */
    public void setBounds(Rectangle r)
    {
        if (_isVisible)
            addDirty(_bounds);
        _bounds.setBounds(r);
        if (_isVisible)
            addDirty(_bounds);
    }
    
    /**
     * @return point sprite's top left corner is at
    */
    public Point getLocation()
    {
        return new Point(_bounds.x, _bounds.y);
    }

    /**
     * @return sprite's x coordinate
     */
    public int getX()
    {
        return _bounds.x;
    }

    /**
     * @return sprite's y coordinate
     */
    public int getY()
    {
        return _bounds.y;
    }
    
    /**
     * @return size of the sprite
    */
    public Dimension getSize()
    {
        return new Dimension(_bounds.width, _bounds.height);
    }

    /**
     * @return sprite's width
     */
    public int getWidth()
    {
        return _bounds.width;
    }

    /**
     * @return sprite's height
     */
    public int getHeight()
    {
        return _bounds.height;
    }
    
    
    /**
     * returns the bounds of the rectangle
    */
    public Rectangle getBounds()
    {
        return new Rectangle(_bounds);
    }
    
    /**
     * Sets if the sprite is visible.
     *
     * @param b true if sprite is visible else false.
    */
    public void setVisible(boolean b)
    {
        _isVisible = b;
        addDirty(_bounds);
    }
    
    /**
     * @return true if sprite is visible, otherwise false
    */
    public boolean isVisible()
    {
        return _isVisible;
    }

    /**
     * Solid sprites fill their bounds with the background color before drawing
     * themselves.
     *
     * @param b true if sprite should be solid
    */
    public void setSolid(boolean b)
    {
        if (b != _isSolid)
            invalidate();
        _isSolid = b;
    }
    
    /**
     * @return true if sprite is solid
    */
    public boolean isSolid()
    {
        return _isSolid;
    }
    
    /**
     * A bounded sprite draws it's bounding box in the bounding color around itself.
     * This is usefull for debugging!
     *
     * @param b set to true if drawing bounding box is desired.
    */
    public void setBounding(boolean b)
    {
        if (_isBounded != b)
            invalidate();
        _isBounded = b;
    }

    /**
     * @return true if bounding box is drawn
    */
    public boolean isBounded()
    {
        return _isBounded;
    }
    
    /**
     * Sets the background color. The background color is only used when a
     * sprite is solid.
     *
     * @param c color of background
    */
    public void setBackground(Color c)
    {
        _background = c;
        if (_isSolid)
            invalidate();
    }
    
    /**
     * Sets the color of the bounding box.
     *
     * @param c Color to set bounding box to
    */
    public void setBoundingColor(Color c)
    {
        _bounding = c;
        if (_isBounded)
            invalidate();
    }
    
    
// **********************************************
// ***** METHODS - Dirty Rectangle handling *****
// **********************************************

    /**
     * adds rectangle to the dirty region
    */
    protected synchronized void addDirty(Rectangle r)
    {
        if (_isDirty)
            _dirty.add(r);
        else
            _dirty.setBounds(r);
        _isDirty = true;
    }
    
    /**
     * @return dirty region or null if not dirty
    */
    public synchronized Rectangle getDirty()
    {
        if (_isDirty == false)
            return null;
        return new Rectangle(_dirty);
    }
    
    /**
     * Marks the sprite as clean
    */
    public synchronized void clean()
    {
        _isDirty =  false;
    }
    
    /**
     * forces the sprite to be dirty. Note that non-visible sprites ignore this.
    */
    public synchronized void invalidate()
    {
        if (_isVisible == false)
            return;
        if (_isDirty)
            _dirty.add(_bounds);
        else
            _dirty.setBounds(_bounds);
        _isDirty = true;
    }
    
    /**
     * Returns if the rectangle is dirty or not
    */
    public synchronized boolean isDirty()
    {
        return _isDirty;
    }


// *************************************
// ***** METHODS - Bounds checking *****
// *************************************
    
    /**
     * @return true if provided point intersects this sprite
    */
    public boolean isIntersction(int x, int y)
    {
        return _bounds.contains(x, y);
    }

    /**
     * @return true if provided point intersects this sprite
    */
    public boolean isIntersection(Point p)
    {
        return _bounds.contains(p);
    }

    /**
     * @return true if provided rectangle intersects this sprite
    */
    public boolean isIntersection(Rectangle r)
    {
        return _bounds.intersects(r);
    }
    
// *****************************
// ***** METHODS - Drawing *****
// *****************************

    /**
     * Draws the sprite on the provided graphics context. It should be noted that sprites ARE NOT required
     * to store the state of any graphics settings so it is possible that the graphics settings will
     * change. Sprite managers should store any settings for the graphics context before rendering the
     * sprites and restore them afterwards. Overriders may want to call this to automatically handle
     * drawing the background and bounding box
     *
     * @param g the graphics context to draw to
     * @param c component required by some awt draw methods
    */
    public void render(Graphics g, Component c)
    {
        if (_isVisible)
        {
            if (_isSolid)
            {
                g.setColor(_background);
                g.fillRect(_bounds.x, _bounds.y, _bounds.width, _bounds.height);
            }
            
            if (_isBounded)
            {
                g.setColor(_bounding);
                g.drawRect(_bounds.x, _bounds.y, _bounds.width, _bounds.height);
            }
        }
        clean();
    }
    
    /**
     * Same as render, but adjusts the sprites coordinates by adding offset to coordinates.
     *
     * @param g the graphics context to draw to
     * @param c component required by some awt draw methods
     * @param offset point relative to top left that sprite is to be drawn
    */
    public void render(Graphics g, Component c, Point offset)
    {
        if (_isVisible)
        {
            if (_isSolid)
            {
                g.setColor(_background);
                g.fillRect(_bounds.x + offset.x, _bounds.y + offset.y, _bounds.width, _bounds.height);
            }
            
            if (_isBounded)
            {
                g.setColor(_bounding);
                g.drawRect(_bounds.x + offset.x, _bounds.y + offset.y, _bounds.width, _bounds.height);
            }
        }
        clean();
    }

}
